# Copyright 2018 Markus Scheidgen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an"AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
import os
import time
import click
import urllib.parse
import requests

from nomad import config
from nomad import processing

from .client import client


def stream_upload_with_client(client, stream, name=None):
    user = client.auth.get_auth().response().result
    token = user.access_token
    url = config.client.url + '/uploads/'
    if name is not None:
        url += '?name=%s' % urllib.parse.quote(name)

    response = requests.put(url, headers={'Authorization': 'Bearer %s' % token}, data=stream)
    if response.status_code != 200:
        raise Exception('nomad return status %d' % response.status_code)
    upload_id = response.json()['upload_id']

    return client.uploads.get_upload(upload_id=upload_id).response().result


def upload_file(file_path: str, name: str = None, offline: bool = False, publish: bool = False, client=None):
    '''
    Upload a file to nomad.

    Arguments:
        file_path: path to the file, absolute or relative to call directory
        name: optional name, default is the file_path's basename
        offline: allows to process data without upload, requires client to be run on the server
        publish: automatically publish after successful processing

    Returns: The upload_id
    '''
    if client is None:
        from nomad.cli.client import create_client
        client = create_client()
    if offline:
        upload = client.uploads.upload(
            local_path=os.path.abspath(file_path), name=name).response().result
        click.echo('process offline: %s' % file_path)
    else:
        # bravado does not seem to support streaming?
        try:
            with open(file_path, 'rb') as f:
                upload = stream_upload_with_client(client, f, name=name)
        except Exception as e:
            click.echo('could not upload the file: %s' % str(e))
            return

    while upload is not None and upload.tasks_status not in [processing.SUCCESS, processing.FAILURE]:
        upload = client.uploads.get_upload(upload_id=upload.upload_id).response().result
        calcs = upload.calcs.pagination
        if calcs is None:
            total, successes, failures = 0, 0, 0
        else:
            total, successes, failures = (calcs.total, calcs.successes, calcs.failures)

        ret = '\n' if upload.tasks_status in (processing.SUCCESS, processing.FAILURE) else '\r'

        print(
            'status: %s; task: %s; parsing: %d/%d/%d                %s' %
            (upload.tasks_status, upload.current_task, successes, failures, total, ret), end='')

        time.sleep(1)

    if upload.tasks_status == processing.FAILURE:
        click.echo('There have been errors:')
        for error in upload.errors:
            click.echo('    %s' % error)
    elif publish:
        client.uploads.exec_upload_operation(upload_id=upload.upload_id, payload=dict(operation='publish')).response()

    return upload.upload_id


@client.command(
    help='Upload files to nomad. The given path can be a single file or a directory. '
    'All .zip files in a directory will be uploaded.')
@click.argument('PATH', nargs=-1, required=True, type=click.Path(exists=True))
@click.option(
    '--name',
    help='Optional name for the upload of a single file. Will be ignored on directories.')
@click.option(
    '--offline', is_flag=True, default=False,
    help='Upload files "offline": files will not be uploaded, but processed were they are. '
    'Only works when run on the nomad host.')
@click.option(
    '--publish', is_flag=True, default=False,
    help='Automatically move upload out of the staging area after successful processing')
def upload(path, name: str, offline: bool, publish: bool):
    paths = path
    click.echo('uploading files from %s paths' % len(paths))
    for path in paths:
        click.echo('uploading %s' % path)
        if os.path.isfile(path):
            name = name if name is not None else os.path.basename(path)
            upload_file(path, name, offline, publish)

        elif os.path.isdir(path):
            for (dirpath, _, filenames) in os.walk(path):
                for filename in filenames:
                    if filename.endswith('.zip'):
                        file_path = os.path.abspath(os.path.join(dirpath, filename))
                        name = os.path.basename(file_path)
                        upload_file(file_path, name, offline, publish)

        else:
            click.echo('Unknown path type %s.' % path)



from __future__ import print_function
from __future__ import division
from past.builtins import cmp
from builtins import zip
from builtins import range
from past.utils import old_div
import readline # optional, will allow Up/Down/History in the console
import code
import re, glob, sys, os, os.path, math, json

def goInteractive(locals):
    """Debugging function that when called stops execution and drops in an interactive loop.

Exiting the interpreter will continue execution.
call as follows:
    goInteractive(locals())
"""
    vars = globals().copy()
    vars.update(locals)
    shell = code.InteractiveConsole(vars)
    shell.interact()

def duLineGBSize(line):
    """returns the size in GB of a line outputted by du -h"""
    sizeRe=re.compile("\s*(?P<nr>[0-9.]+)(?P<unit>[kMGT]?)")
    m = sizeRe.match(line)
    if m:
        unit = {
            "": 1.0/(1024*1024*1024),
            "K": 1.0/(1024*1024),
            "M":1.0/1024,
            "G":1.0,
            "T":1024.0
        }
        return float(m.group("nr"))*unit[m.group("unit")]
    else:
        return 0.0

def totalGBSize(filePath):
    """returns the total size of the contents of a file containing du -h lines

    Makes a simple sum, which is wrong if both the directories and their contents
    are listed"""
    totS = 0.0
    with open(filePath) as f:
        while True:
            line = f.readline()
            if not line: break
            totS += duLineGBSize(line)
    return totS

def makeSplits(filePath, outSplitsF):
    groups = readDuFileToGroups(filePath)
    splits = splitGroups(groups)
    printSplits(splits)
    writeSplits(splits, outSplitsF)

def readDuFileToGroups(filePath, basePath = None):
    if basePath is None:
        basePath = os.getcwd()
    groups={}
    duLineRe = re.compile(r"\s*(?P<size>[0-9.]+[KMGT]?)\s+(?P<dir>.+)$")
    with open(filePath) as f:
        while True:
            line = f.readline()
            if not line: break
            if not line.isspace():
                m = duLineRe.match(line)
                if not m:
                    print("failed to match", repr(line))
                directory = os.path.normpath(os.path.join(basePath, m.group("dir")))
                group = os.path.dirname(directory)
                dirInGroup = groups.get(group, None)
                if dirInGroup:
                    dirInGroup[directory] = duLineGBSize(line)
                else:
                    groups[group] = {
                        directory: duLineGBSize(line)
                    }
    return groups

def splitGroups(groups, joinSmall = True):
    smallGroupsSizes = {}
    splits = []
    for group, dirInGroup in groups.items():
        groupSize = sum(dirInGroup.values())
        if joinSmall and groupSize < 0.5:
            parent = os.path.dirname(group)
            smallInParent = smallGroupsSizes.get(parent, None)
            if smallInParent:
                smallInParent[group] = groupSize
            else:
                smallGroupsSizes[parent] = {
                    group: groupSize
                }
        elif groupSize < 16:
            contents = {}
            for dirPath, size in dirInGroup.items():
                contents[os.path.basename(dirPath)] = size
            splits.append({
                "baseDir": group,
                "contents": contents,
                "size": groupSize
            })
        else:
            items = list(dirInGroup.items())
            items.sort(lambda x, y: cmp(x[1], y[1]))
            nSplits = int(math.ceil(groupSize / 14))
            splitsEls = []
            for i in range(nSplits):
                splitsEls.append({})
            while items:
                for i in range(nSplits):
                    if not items: break
                    k, v = items.pop()
                    splitsEls[i][k] = v
            for split in splitsEls:
                contents = {}
                for dirPath, size in split.items():
                    contents[os.path.basename(dirPath)] = size
                splits.append({
                        "baseDir": group,
                        "contents": contents,
                        "size": sum(split.values())
                    })
    if len(smallGroupsSizes) == 0:
        pass
    else:
        splits.extend(splitGroups(smallGroupsSizes, joinSmall = False))
    return splits

def printSplits(splits, fOut = sys.stdout):
    splits = list(splits)
    splits.sort(lambda x, y: cmp(x["size"], y["size"]))
    sizes = [x.get("size", 0) for x in splits]
    sumSize = sum(sizes)
    json.dump(splits, fOut, indent = 2)
    fOut.write("\nmin: %6.3f max: %6.3f avg: %6.3f (%7.3f for %d blocks)\n" %
               (sizes[0], sizes[-1], sumSize/len(splits), sumSize, len(splits)))

def writeSplits(splits, fOut = sys.stdout):
    for split in splits:
        fOut.write("\n")
        fOut.write(split["baseDir"])
        fOut.write("\n")
        contents = split.get("contents")
        if contents:
            for c in contents.keys():
                fOut.write(c)
                fOut.write("\n")


def filesGbSizes(globPattern):
    splitSizes = {}
    for f in glob.glob(globPattern):
        splitSizes[f] = totalGBSize(f)
    return splitSizes

def writeFilesGbSizes(globPattern):
    "prints the sizes of the files matching the glob pattern"""
    minSize = None
    maxSize = 0.0
    sumSize = 0.0
    nSizes = 0
    fileSizes = list(filesGbSizes(globPattern).items())
    fileSizes.sort(lambda x,y: cmp(x[1],y[1]))
    for f, sizeNow in fileSizes:
        print("%6.3f %s" % (sizeNow , f))
        if minSize is None or minSize > sizeNow:
            minSize = sizeNow
        if maxSize < sizeNow:
            maxSize = sizeNow
        nSizes += 1
        sumSize += sizeNow
    print()
    print("min: %6.3f max: %6.3f avg: %6.3f (%7.3f for %d blocks)" % (
        minSize, maxSize, sumSize / nSizes, sumSize, nSizes))

def writeOut(outF):
    for f in glob.glob("aflowLib3splits*"):
        outF.write("/aflowlib_data/LIB3_LIB/\n")
        for line in open(f).readlines():
            cont = transformLine(line)
            if cont:
                outF.write(cont)
                outF.write("\n")
        outF.write("\n")

lineRe=re.compile("[0-9.]+[KMGT]?\t.*/([^/]+)\n$")
lineRe=re.compile("[0-9.]+[KMGT]?\tLIB3_LIB/([^/]+)\n$")
def transformLine(line):
    m = lineRe.match(line)
    if m:
        return m.groups()[0]
    else:
        if not line.isspace():
            raise Exception("invalid line %s" % line)
        return None

def oqmdSplit(oqmdSplitFile, filesToWrite = None):
    if filesToWrite is None:
        filesToWrite = ["bryce", "perovskite", "thermoelectrics", "elements", "voorhees", "garnet", "ilmenite", "scott", "spinel", "cod", "heusler", "james"]
    # "icsd", "enum_all"
    fs = {}
    s=re.compile("/")
    r=re.compile(r"[0-9.]+[KMGT]?\s\./ftp_upload_for_uid_290/oqmd/(?P<topLevel>[^/]+).*")
    for f in filesToWrite:
        fs[f] = open("ftp_upload_for_uid_290." + f, "w")
    depth = { "binaries": 5, "icsd": 4, "bryce":5, "perovskite":5, "thermoelectrics": 5, "elements": 5, "voorhees": 4, "enum_all": 7, "garnet":5, "ilmenite":4, "scott":4, "spinel": 5, "cod": 4, "cod": 4, "heusler": 5, "james":5 }
    while True:
        line = oqmdSplitFile.readline()
        if not line: break
        m = r.match(line)
        if m:
            topLevel=m.group("topLevel")
            outF = fs.get(topLevel, None)
            if outF and depth[topLevel] == len(s.findall(line)):
                outF.write(line)

def readSizes(duFilePath):
    "return the sizes contained in a du file"
    allSizes = []
    with open(duFilePath) as fIn:
        while True:
            line = fIn.readline()
            if not line: break
            size = duLineGBSize(line)
            allSizes.append(size)
    return allSizes

def binGBSizes(allSizes):
    """returns the sizes binned in logarithmic spaced bins"""
    bins = [(1024**-2,"1K"),(10*1024**-2,"10K"),(100*1024**-2,"100K"),(1024**-1,"1M"),(10*1024**-1,"10M"), (100*1024**-1,"100M"),(1,"1G"),(10,"10G"),(100,"100G"),(1024,"1T"),(10*1024,"10T")]
    binnedValues = [ 0 for i in range(len(bins))]
    cumulativeSizes = [0.0 for i in range(len(bins))]
    for value in allSizes:
        i = 0
        while value > bins[i][0]:
            i += 1
        binnedValues[i] += 1
        cumulativeSizes[i] += value
    return { "binnedValues": binnedValues, "cumulativeValue": cumulativeSizes, "bins": bins }

def binPairValues(values, bins):
    binnedValues = [ 0 for i in range(len(bins))]
    for value, nVal in values:
        i = 0
        while value > bins[i]:
            i += 1
        binnedValues[i] += nVal
    return { "binnedValues": binnedValues, "bins": bins }

kindDist=[u'nElementalMetalsR', u'nBinaryAlloyR' ,  u'nSemiconductorsR', u'nMetalsR' , u'nRuns']

def printOut(dict):
    for k, v in dict.items():
        sys.stdout.write("%s\t%s\n" % (k, v))

bins = [3.333,10,33.33,100,333.3,1000,3333,10000, 33333, 100000]

def toBinPairs(hist):
    return [(int(x[0]), x[1]) for x in list(hist.items())]

def printHist(histDict, outF = sys.stdout):
    res=binPairValues(toBinPairs(histDict), bins)
    for b,v in zip(res["bins"], res["binnedValues"]):
        outF.write("%s\t%s\n" % (b, v))

def printTopValues(topValues, fOut = sys.stdout):
    for n,l in topValues:
        fOut.write("%s\t%s\n" %(" ".join(l), n))



import numpy as np            # pylint: disable=unused-import
import typing                 # pylint: disable=unused-import
from nomad.metainfo import (  # pylint: disable=unused-import
    MSection, MCategory, Category, Package, Quantity, Section, SubSection, SectionProxy,
    Reference
)
from nomad.metainfo.legacy import LegacyDefinition

from nomad.datamodel.metainfo import common
from nomad.datamodel.metainfo import public

m_package = Package(
    name='orca_nomadmetainfo_json',
    description='None',
    a_legacy=LegacyDefinition(name='orca.nomadmetainfo.json'))


class x_orca_atom_positions(MSection):
    '''
    positions of an atom
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_orca_atom_positions'))

    x_orca_atom_labels = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_atom_labels'))

    x_orca_atom_positions_x = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='meter',
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_atom_positions_x'))

    x_orca_atom_positions_y = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='meter',
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_atom_positions_y'))

    x_orca_atom_positions_z = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='meter',
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_atom_positions_z'))


class x_orca_final_geometry(MSection):
    '''
    final optimized positions of an atom
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_orca_final_geometry'))

    x_orca_atom_labels_geo_opt = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_atom_labels_geo_opt'))

    x_orca_atom_positions_x_geo_opt = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='meter',
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_atom_positions_x_geo_opt'))

    x_orca_atom_positions_y_geo_opt = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='meter',
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_atom_positions_y_geo_opt'))

    x_orca_atom_positions_z_geo_opt = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='meter',
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_atom_positions_z_geo_opt'))


class x_orca_section_functionals(MSection):
    '''
    XC functional
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_orca_section_functionals'))


class section_method(public.section_method):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_method'))

    x_orca_1_elect_energy_change = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_1_elect_energy_change'))

    x_orca_angular_grid = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_angular_grid'))

    x_orca_atomic_orbital_integral_source = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_atomic_orbital_integral_source'))

    x_orca_avg_nb_grid_pts_per_atom = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_avg_nb_grid_pts_per_atom'))

    x_orca_avg_nb_points_per_batch = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_avg_nb_points_per_batch'))

    x_orca_basis_fn_cutoff = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_basis_fn_cutoff'))

    x_orca_beckes_beta_param = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_beckes_beta_param'))

    x_orca_Brueckner_orbitals_calc_on_off = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_Brueckner_orbitals_calc_on_off'))

    x_orca_ccsdt_aaa_triples_contribution = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_ccsdt_aaa_triples_contribution'))

    x_orca_ccsdt_aab_triples_contribution = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_ccsdt_aab_triples_contribution'))

    x_orca_ci_half_s_and_s_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_ci_half_s_and_s_energy'))

    x_orca_ci_iteration_nb = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_ci_iteration_nb'))

    x_orca_convergence_check_mode = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_convergence_check_mode'))

    x_orca_convergence_tol_max_residuum = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_convergence_tol_max_residuum'))

    x_orca_correl_functional = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_correl_functional'))

    x_orca_coulomb_transformation_dimension_basis = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_coulomb_transformation_dimension_basis'))

    x_orca_coulomb_transformation_type = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_coulomb_transformation_type'))

    x_orca_energy_change_tolerance = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_energy_change_tolerance'))

    x_orca_exchange_functional = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_exchange_functional'))

    x_orca_f12_correction_on_off = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_f12_correction_on_off'))

    x_orca_frozen_core_treatment = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_frozen_core_treatment'))

    x_orca_gral_integ_accuracy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_gral_integ_accuracy'))

    x_orca_grid_pruning_method = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_grid_pruning_method'))

    x_orca_hf_method = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_hf_method'))

    x_orca_hf_type = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_hf_type'))

    x_orca_integr_weight_cutoff = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_integr_weight_cutoff'))

    x_orca_integral_package_used = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_integral_package_used'))

    x_orca_integral_transformation = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_integral_transformation'))

    x_orca_K_C_formation = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_K_C_formation'))

    x_orca_lda_part_of_gga_corr = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_lda_part_of_gga_corr'))

    x_orca_level_shift_amplitude_update = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_level_shift_amplitude_update'))

    x_orca_mp2_aux_basis_dimension = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_mp2_aux_basis_dimension'))

    x_orca_mp2_basis_dimension = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_mp2_basis_dimension'))

    x_orca_mp2_initial_guess = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_mp2_initial_guess'))

    x_orca_multiplicity = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_multiplicity'))

    x_orca_nb_alpha_pairs_included = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_nb_alpha_pairs_included'))

    x_orca_nb_beta_pairs_included = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_nb_beta_pairs_included'))

    x_orca_nb_grid_pts_after_initial_pruning = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_nb_grid_pts_after_initial_pruning'))

    x_orca_nb_grid_pts_after_weights_screening = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_nb_grid_pts_after_weights_screening'))

    x_orca_nb_internal_alpha_mol_orbitals = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_nb_internal_alpha_mol_orbitals'))

    x_orca_nb_internal_beta_mol_orbitals = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_nb_internal_beta_mol_orbitals'))

    x_orca_nb_of_atomic_orbitals = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_nb_of_atomic_orbitals'))

    x_orca_nb_of_correlated_electrons = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_nb_of_correlated_electrons'))

    x_orca_nb_of_electrons = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_nb_of_electrons'))

    x_orca_nelectrons = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_nelectrons'))

    x_orca_nuclear_repulsion = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_nuclear_repulsion'))

    x_orca_orbital_opt_on_off = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_orbital_opt_on_off'))

    x_orca_pair_cutoff = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_pair_cutoff'))

    x_orca_perturbative_triple_excitations_on_off = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_perturbative_triple_excitations_on_off'))

    x_orca_radial_grid_type = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_radial_grid_type'))

    x_orca_reference_wave_function = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_reference_wave_function'))

    x_orca_scalar_relativistic_method = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_scalar_relativistic_method'))

    x_orca_scaling_mp2_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_scaling_mp2_energy'))

    x_orca_single_excitations_on_off = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_single_excitations_on_off'))

    x_orca_single_norm_half_ss = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_single_norm_half_ss'))

    x_orca_speed_of_light_used = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_speed_of_light_used'))

    x_orca_total_charge = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_total_charge'))

    x_orca_total_nb_batches = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_total_nb_batches'))

    x_orca_total_nb_grid_pts = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_total_nb_grid_pts'))

    x_orca_total_nb_pairs_included = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_total_nb_pairs_included'))

    x_orca_wave_function_correlation_treatment = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_wave_function_correlation_treatment'))

    x_orca_weight_gener_scheme = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_weight_gener_scheme'))

    x_orca_xalpha_param = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_xalpha_param'))

    x_orca_XC_functional_type = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_XC_functional_type'))

    x_orca_z_vector_calc_on_off = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_z_vector_calc_on_off'))

    x_orca_section_functionals = SubSection(
        sub_section=SectionProxy('x_orca_section_functionals'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_orca_section_functionals'))


class section_scf_iteration(public.section_scf_iteration):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_scf_iteration'))

    x_orca_angular_grid_final = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_angular_grid_final'))

    x_orca_avg_nb_grid_pts_per_atom_final = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_avg_nb_grid_pts_per_atom_final'))

    x_orca_avg_nb_points_per_batch_final = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_avg_nb_points_per_batch_final'))

    x_orca_basis_fn_cutoff_final = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_basis_fn_cutoff_final'))

    x_orca_gral_integ_accuracy_final = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_gral_integ_accuracy_final'))

    x_orca_grid_pruning_method_final = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_grid_pruning_method_final'))

    x_orca_integr_weight_cutoff_final = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_integr_weight_cutoff_final'))

    x_orca_iteration_nb = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_iteration_nb'))

    x_orca_last_energy_change_tolerance = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_last_energy_change_tolerance'))

    x_orca_last_energy_change = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_last_energy_change'))

    x_orca_last_max_density_change = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_last_max_density_change'))

    x_orca_last_max_density_tolerance = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_last_max_density_tolerance'))

    x_orca_last_rms_density_change = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_last_rms_density_change'))

    x_orca_last_rms_density_tolerance = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_last_rms_density_tolerance'))

    x_orca_nb_grid_pts_after_initial_pruning_final = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_nb_grid_pts_after_initial_pruning_final'))

    x_orca_nb_grid_pts_after_weights_screening_final = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_nb_grid_pts_after_weights_screening_final'))

    x_orca_radial_grid_type_final = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_radial_grid_type_final'))

    x_orca_total_nb_batches_final = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_total_nb_batches_final'))

    x_orca_total_nb_grid_pts_final = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_total_nb_grid_pts_final'))

    x_orca_weight_gener_scheme_final = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_weight_gener_scheme_final'))


class section_dos(public.section_dos):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_dos'))

    x_orca_atom_nb_mroc = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_atom_nb_mroc'))

    x_orca_atom_nb = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_atom_nb'))

    x_orca_atom_orbital_mroc = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_atom_orbital_mroc'))

    x_orca_atom_species_mroc = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_atom_species_mroc'))

    x_orca_atom_species = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_atom_species'))

    x_orca_mulliken_atom_charge = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_mulliken_atom_charge'))

    x_orca_mulliken_partial_orbital_charge_mroc = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_mulliken_partial_orbital_charge_mroc'))

    x_orca_mulliken_total_charge = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_mulliken_total_charge'))


class section_system(public.section_system):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_system'))

    x_orca_atom_positions = SubSection(
        sub_section=SectionProxy('x_orca_atom_positions'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_orca_atom_positions'))


class section_basis_set(public.section_basis_set):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_basis_set'))

    x_orca_auxiliary_basis_set_contracted = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_auxiliary_basis_set_contracted'))

    x_orca_auxiliary_basis_set = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_auxiliary_basis_set'))

    x_orca_basis_set_atom_labels = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_basis_set_atom_labels'))

    x_orca_basis_set_contracted = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_basis_set_contracted'))

    x_orca_basis_set = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_basis_set'))

    x_orca_highest_angular_moment_aux = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_highest_angular_moment_aux'))

    x_orca_highest_angular_moment = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_highest_angular_moment'))

    x_orca_maximum_contraction_depth_aux = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_maximum_contraction_depth_aux'))

    x_orca_maximum_contraction_depth = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_maximum_contraction_depth'))

    x_orca_nb_of_contracted_basis_functions_aux = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_nb_of_contracted_basis_functions_aux'))

    x_orca_nb_of_contracted_basis_functions = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_nb_of_contracted_basis_functions'))

    x_orca_nb_of_contracted_shells_aux = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_nb_of_contracted_shells_aux'))

    x_orca_nb_of_contracted_shells = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_nb_of_contracted_shells'))

    x_orca_nb_of_primitive_gaussian_functions_aux = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_nb_of_primitive_gaussian_functions_aux'))

    x_orca_nb_of_primitive_gaussian_functions = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_nb_of_primitive_gaussian_functions'))

    x_orca_nb_of_primitive_gaussian_shells_aux = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_nb_of_primitive_gaussian_shells_aux'))

    x_orca_nb_of_primitive_gaussian_shells = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_nb_of_primitive_gaussian_shells'))

    x_orca_nb_primitive_gaussian_shells_aux = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_nb_primitive_gaussian_shells_aux'))

    x_orca_nb_primitive_gaussian_shells = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_nb_primitive_gaussian_shells'))


class section_single_configuration_calculation(public.section_single_configuration_calculation):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_single_configuration_calculation'))

    x_orca_basis_fn_evaluation = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_basis_fn_evaluation'))

    x_orca_ccsd_correlation_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_ccsd_correlation_energy'))

    x_orca_ccsd_final_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_ccsd_final_energy'))

    x_orca_ccsd_t_final_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_ccsd_t_final_energy'))

    x_orca_ccsd_total_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_ccsd_total_energy'))

    x_orca_ccsdt_aba_triples_contribution = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_ccsdt_aba_triples_contribution'))

    x_orca_ccsdt_bbb_triples_contribution = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_ccsdt_bbb_triples_contribution'))

    x_orca_ccsdt_final_corr_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_ccsdt_final_corr_energy'))

    x_orca_ccsdt_total_triples_correction = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_ccsdt_total_triples_correction'))

    x_orca_ci_correl_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_ci_correl_energy'))

    x_orca_ci_deltaE_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_ci_deltaE_energy'))

    x_orca_ci_iteration_time = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_ci_iteration_time'))

    x_orca_ci_residual_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_ci_residual_energy'))

    x_orca_ci_total_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_ci_total_energy'))

    x_orca_correlation_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_correlation_energy'))

    x_orca_coulomb_formation = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_coulomb_formation'))

    x_orca_density_evaluation = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_density_evaluation'))

    x_orca_density_matrix_formation = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_density_matrix_formation'))

    x_orca_diagonalization = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_diagonalization'))

    x_orca_diis_solution = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_diis_solution'))

    x_orca_elec_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_elec_energy'))

    x_orca_exchange_correlation_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_exchange_correlation_energy'))

    x_orca_exchange_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_exchange_energy'))

    x_orca_final_time = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_final_time'))

    x_orca_fock_matrix_formation = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_fock_matrix_formation'))

    x_orca_geo_opt_cycle = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_geo_opt_cycle'))

    x_orca_grid_generation = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_grid_generation'))

    x_orca_initial_guess = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_initial_guess'))

    x_orca_kinetc_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_kinetc_energy'))

    x_orca_mp2_corr_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_mp2_corr_energy'))

    x_orca_mp2_energy_spin_aa = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_mp2_energy_spin_aa'))

    x_orca_mp2_energy_spin_ab = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_mp2_energy_spin_ab'))

    x_orca_mp2_energy_spin_bb = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_mp2_energy_spin_bb'))

    x_orca_mp2_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_mp2_energy'))

    x_orca_mp2_total_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_mp2_total_energy'))

    x_orca_nb_elect_alpha_channel = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_nb_elect_alpha_channel'))

    x_orca_nb_elect_beta_channel = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_nb_elect_beta_channel'))

    x_orca_nb_elect_total = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_nb_elect_total'))

    x_orca_nuc_repulsion = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_nuc_repulsion'))

    x_orca_one_elec_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_one_elec_energy'))

    x_orca_orbital_orthonormalization = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_orbital_orthonormalization'))

    x_orca_orbital_transformation = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_orbital_transformation'))

    x_orca_population_analysis = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_population_analysis'))

    x_orca_potential_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_potential_energy'))

    x_orca_potential_evaluation = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_potential_evaluation'))

    x_orca_split_rj = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_split_rj'))

    x_orca_sum_individual_times = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_sum_individual_times'))

    x_orca_t1_diagnostic = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_t1_diagnostic'))

    x_orca_T_and_T_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_T_and_T_energy'))

    x_orca_total_days_time = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_total_days_time'))

    x_orca_total_hours_time = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_total_hours_time'))

    x_orca_total_mins_time = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_total_mins_time'))

    x_orca_total_secs_time = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_total_secs_time'))

    x_orca_two_elec_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_two_elec_energy'))

    x_orca_virial_ratio = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_virial_ratio'))

    x_orca_xc_functional_evaluation = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_xc_functional_evaluation'))

    x_orca_xc_integration = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_xc_integration'))


class section_sampling_method(public.section_sampling_method):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_sampling_method'))

    x_orca_coords_choice_name = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_coords_choice_name'))

    x_orca_coords_choice = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_coords_choice'))

    x_orca_energy_change_tol_value = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_energy_change_tol_value'))

    x_orca_energy_change_tol = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_energy_change_tol'))

    x_orca_initial_hessian_name = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_initial_hessian_name'))

    x_orca_initial_hessian = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_initial_hessian'))

    x_orca_max_displacement_tol_value = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_max_displacement_tol_value'))

    x_orca_max_displacement_tol = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_max_displacement_tol'))

    x_orca_max_gradient_tol_value = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_max_gradient_tol_value'))

    x_orca_max_gradient_tol = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_max_gradient_tol'))

    x_orca_rms_displacement_tol_value = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_rms_displacement_tol_value'))

    x_orca_rms_displacement_tol = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_rms_displacement_tol'))

    x_orca_rms_gradient_tol_value = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_rms_gradient_tol_value'))

    x_orca_rms_gradient_tol = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_rms_gradient_tol'))

    x_orca_update_method_name = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_update_method_name'))

    x_orca_update_method = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_update_method'))

    x_orca_final_geometry = SubSection(
        sub_section=SectionProxy('x_orca_final_geometry'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_orca_final_geometry'))


class section_excited_states(common.section_excited_states):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_excited_states'))

    x_orca_excitation_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_excitation_energy'))

    x_orca_oscillator_strength = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_oscillator_strength'))

    x_orca_transition_dipole_moment_x = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_transition_dipole_moment_x'))

    x_orca_transition_dipole_moment_y = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_transition_dipole_moment_y'))

    x_orca_transition_dipole_moment_z = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_transition_dipole_moment_z'))


class section_eigenvalues(public.section_eigenvalues):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_eigenvalues'))

    x_orca_orbital_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_orbital_energy'))

    x_orca_orbital_nb = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_orbital_nb'))

    x_orca_orbital_occupation_nb = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_orbital_occupation_nb'))


class section_run(public.section_run):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_run'))

    x_orca_program_compilation_date = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_program_compilation_date'))

    x_orca_program_compilation_time = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_program_compilation_time'))

    x_orca_program_svn = Quantity(
        type=str,
        shape=[],
        description='''
        svn revision of the program
        ''',
        a_legacy=LegacyDefinition(name='x_orca_program_svn'))

    x_orca_program_version = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_orca_program_version'))


m_package.__init_metainfo__()



# Copyright 2016-2018 Martina Stella, Fawzi Mohamed
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from __future__ import division
from builtins import range
from builtins import object
import numpy as np
import nomadcore.ActivateLogging
from nomadcore.caching_backend import CachingLevel
from nomadcore.simple_parser import mainFunction
from nomadcore.simple_parser import SimpleMatcher as SM
from nomadcore.local_meta_info import loadJsonFile, InfoKindEl
from .OnetepCommon import get_metaInfo
import logging, os, re, sys



############################################################
# This is the parser for the *.md file of onetep.
############################################################

logger = logging.getLogger("nomad.OnetepMDParser")

class OnetepMDParserContext(object):
    """Context for parsing Onetep *.md file.


    The onClose_ functions allow processing and writing of cached values after a section is closed.
    They take the following arguments:
        backend: Class that takes care of wrting and caching of metadata.
        gIndex: Index of the section that is closed.
        section: The cached values and sections that were found in the section that is closed.
    """
    def __init__(self, writeMetaData = True):
        """Args:
            writeMetaData: Deteremines if metadata is written or stored in class attributes.
        """
        self.writeMetaData = writeMetaData
        self.frame_temperature =[]
        self.frame_pressure = []
        self.total_energy = []
        self.hamiltonian = []
        self.kinetic = []
        self.md_forces =[]
        self.md_veloc =[]
        self.frame_atom_label =[]
        self.total_velocities =[]
        self.total_forces =[]
        self.atom_label = []
        self.frame_stress_tensor =[]
        self.total_positions =[]
        self.frame_cell=[]
        self.vector_velocities=[]
        self.frame_time =[]
    def startedParsing(self, fInName, parser):
        """Function is called when the parsing starts and the compiled parser is obtained.

        Args:
            fInName: The file name on which the current parser is running.
            parser: The compiled parser. Is an object of the class SimpleParser in nomadcore.simple_parser.py.
        """
        self.parser = parser

    def onClose_x_onetep_section_md(self, backend, gIndex, section):
        temp = section ['x_onetep_md_temperature']
        vet_veloc = section ['x_onetep_md_cell_vectors_vel']
        velocities = section ['x_onetep_md_veloc']
        vet = section ['x_onetep_md_cell_vectors']
        forces = section ['x_onetep_md_forces']
        atoms_lab = section ['x_onetep_md_lab']
        position = section ['x_onetep_md_positions']
        press = section ['x_onetep_md_pressure']
        energies = section['x_onetep_md_energies']
        stress_tensor = section ['x_onetep_md_stress_tensor']

        Hr_J_converter = float(4.35974e-18)
        HrK_to_K_coverter= float(3.1668114e-6)

        if temp:
            for i in range(len(temp)):
                temp[i] = temp[i]/HrK_to_K_coverter

            self.frame_temperature.append(temp[i])

        if press:
            for i in range(len(press)):
                press[i] = press[i] / 10e9
            self.frame_pressure.append(press[i])

        for i in range(len(temp)):
            energies[i] = energies[i].split()
            energies[i] = [float(j) for j in energies[i]]
            energy_list = energies[i]
            energy_list = [x * Hr_J_converter for x in energy_list]
            self.total_energy.append(energy_list[0])
            self.hamiltonian.append(energy_list[1])
            self.kinetic.append(energy_list[2])

        if vet:
            self.cell =[]
            for i in range(len(vet)):
                vet[i] = vet[i].split()
                vet[i] = [float(j) for j in vet[i]]
                vet_list = vet[i]
                self.cell.append(vet_list)
        self.frame_cell.append(self.cell)


        if vet_veloc:
            self.vet_vel =[]
            for i in range(len(vet_veloc)):
                vet_veloc[i] = vet_veloc[i].split()
                vet_veloc[i] = [float(k) for k in vet_veloc[i]]
                v_vet = vet_veloc[i]
                self.vet_vel.append(v_vet)
            self.vector_velocities.append(self.vet_vel)

        if stress_tensor is not None:
            self.stress_tensor_value =[]
            for s in stress_tensor:
                s = s.split()
                s = [float(k) for k in s]
                stress_tens_int = s
                stress_tens_int = [x / 10e9 for x in stress_tens_int]
                self.stress_tensor_value.append(stress_tens_int)
            self.frame_stress_tensor.append(self.stress_tensor_value)

        if position:
            self.at_nr = len(position)
            self.atom_position=[]
            for i in range(0, self.at_nr):
                position[i] = position[i].split()
                position[i] = [float(j) for j in position[i]]
                pos_list = position[i]
                self.atom_position.append(pos_list)
            self.total_positions.append(self.atom_position)


        if velocities is not None:
            self.md_veloc =[]
            for j in range(len(velocities)):
                velocities[j] = velocities[j].split()
                velocities[j] = [float(k) for k in velocities[j]]
                v_st_int = velocities[j]
                self.md_veloc.append(v_st_int)
            self.total_velocities.append(self.md_veloc)

        if forces is not None:
            print (forces,'ciao')
            self.md_forces = []
            for f in forces:
                f = f.split()
                f = [float(k) for k in f]
                f_st_int = f
                self.md_forces.append(f_st_int)
            self.total_forces.append(self.md_forces)







def build_OnetepMDFileSimpleMatcher():
    """Builds the SimpleMatcher to parse the *.md file of onetep.

    SimpleMatchers are called with 'SM (' as this string has length 4,
    which allows nice formating of nested SimpleMatchers in python.

    Returns:
       SimpleMatcher that parses *.md file of onetep.
    """
    return SM (name = 'Root1',
        startReStr = "",
        sections = ['section_run'],
        forwardMatch = True,
        weak = True,
        subMatchers = [
            SM (name = 'Root2',
            startReStr =r"\s*(?P<x_onetep_md_energies>[-+0-9.eEdD]+\s*[-+0-9.eEdD]+\s*[-+0-9.eEdD]+)\s*\<\-\-\sE\s*",
            endReStr ="/n",
            sections = ['x_onetep_section_md'],
            repeats = True,
            subMatchers = [
                SM (r"\s*(?P<x_onetep_md_temperature>[-+0-9.eEdD]+)\s*\<\-\-\sT\s*"),
                SM (r"\s*(?P<x_onetep_md_pressure>[-+0-9.eEdD]+)\s*\<\-\-\sP\s*"),
                SM (r"\s*(?P<x_onetep_md_cell_vectors>[-+0-9.eEdD]+\s*[-+0-9.eEdD]+\s*[-+0-9.eEdD]+)\s*\<\-\-\sh\s",repeats = True),
                SM (r"\s*(?P<x_onetep_md_cell_vectors_vel>[-+0-9.eEdD]+\s*[-+0-9.eEdD]+\s*[-+0-9.eEdD]+)\s*\<\-\-\sh[a-z]\s*",repeats = True),
                SM (r"\s*(?P<x_onetep_md_stress_tensor>[-+0-9.eEdD]+\s*[-+0-9.eEdD]+\s*[-+0-9.eEdD]+)\s*\<\-\-\sS\s",repeats = True),
                SM(r"\s(?P<x_onetep_md_lab>[A-Za-z]+\s*[0-9.]+)\s*(?P<x_onetep_md_positions>[-+0-9.eEdD]+\s*[-+0-9.eEdD]+\s*[-+0-9.eEdD]+)\s*\<\-\-\sR\s*",repeats = True),
                SM(r"\s[A-Za-z]+\s*[0-9.]+\s*(?P<x_onetep_md_veloc>[-+0-9.eEdD]+\s*[-+0-9.eEdD]+\s*[-+0-9.eEdD]+)\s*\<\-\-\sV\s*",repeats = True),
                SM(r"\s[A-Za-z]+\s*[0-9.]+\s*(?P<x_onetep_md_forces>[-+0-9.eEdD]+\s*[-+0-9.eEdD]+\s*[-+0-9.eEdD]+)\s*\<\-\-\sF\s*",repeats = True,endReStr ="/n"),

            ]),
        ])






def get_cachingLevelForMetaName(metaInfoEnv, CachingLvl):
    """Sets the caching level for the metadata.

    Args:
        metaInfoEnv: metadata which is an object of the class InfoKindEnv in nomadcore.local_meta_info.py.
        CachingLvl: Sets the CachingLevel for the sections k_band, run, and single_configuration_calculation.
            This allows to run the parser without opening new sections.

    Returns:
        Dictionary with metaname as key and caching level as value.
    """
    # manually adjust caching of metadata
    cachingLevelForMetaName = { 'x_onetep_md_energies':CachingLevel.Cache,
                                'x_onetep_md_forces':CachingLevel.Cache,
                                'x_onetep_md_veloc':CachingLevel.Cache,
                                'x_onetep_md_lab':CachingLevel.Cache,
                                'x_onetep_md_positions':CachingLevel.Cache,
                                'x_onetep_md_temperature':CachingLevel.Cache,
                                'x_onetep_md_pressure':CachingLevel.Cache,
                                'x_onetep_md_cell_vectors_vel':CachingLevel.Cache,
                                'x_onetep_md_cell_vectors':CachingLevel.Cache,
                                'x_onetep_md_stress_tensor':CachingLevel.Cache,


                                'section_run': CachingLvl,
                                'x_onetep_section_md': CachingLvl,
                               # 'section_single_configuration_calculation': CachingLvl,
                              }
    # Set all band metadata to Cache as they need post-processsing.
    # for name in metaInfoEnv.infoKinds:
    #     if name.startswith('Onetep_'):
    #         cachingLevelForMetaName[name] = CachingLevel.Cache
    return cachingLevelForMetaName


def main(CachingLvl):
    """Main function.

    Set up everything for the parsing of the onetep *.cell file and run the parsing.

    Args:
        CachingLvl: Sets the CachingLevel for the sections k_band, run, and single_configuration_calculation.
            This allows to run the parser without opening new sections.
    """
    # get band.out file description
    OnetepMDFileSimpleMatcher = build_OnetepMDFileSimpleMatcher()
    # loading metadata from nomad-meta-info/meta_info/nomad_meta_info/onetep.nomadmetainfo.json
    metaInfoPath = os.path.normpath(os.path.join(os.path.dirname(os.path.abspath(__file__)),"../../../../nomad-meta-info/meta_info/nomad_meta_info/onetep.nomadmetainfo.json"))
    metaInfoEnv = get_metaInfo(metaInfoPath)
    # set parser info
    parserInfo = {'name':'onetep-md-parser', 'version': '1.0'}
    # get caching level for metadata
    cachingLevelForMetaName = get_cachingLevelForMetaName(metaInfoEnv, CachingLvl)
    # start parsing
    mainFunction(mainFileDescription = OnetepMDFileSimpleMatcher,
                 metaInfoEnv = metaInfoEnv,
                 parserInfo = parserInfo,
                 cachingLevelForMetaName = cachingLevelForMetaName,
                 superContext = OnetepMDParserContext())

if __name__ == "__main__":
    main(CachingLevel.Forward)










# Copyright 2016-2018 Daria M. Tomecka, Fawzi Mohamed
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

################################################################
# This is the parser for the main output file of Fleur (out) - adapted for version fleur26e
################################################################

from builtins import object
from nomadcore.simple_parser import SimpleMatcher as SM
from nomadcore.simple_parser import mainFunction, AncillaryParser, CachingLevel
from nomadcore.local_meta_info import loadJsonFile, InfoKindEl
import os, sys, json, logging
import numpy as np
#import fleur_parser_inp
#import fleur_XML_parser


__author__ = "Daria M. Tomecka"
__maintainer__ = "Daria M. Tomecka"
__email__ = "tomeckadm@gmail.com;"
__date__ = "15/05/2017"

class FleurContext(object):
    """context for the fleur parser"""

    def __init__(self):
        self.parser = None
        self.rootSecMethodIndex = None
        self.secMethodIndex = None
        self.secSystemIndex = None

    def initialize_values(self):
        """allows to reset values if the same superContext is used to parse different files"""
       # pass

#        self.metaInfoEnv = self.parser.parserBuilder.metaInfoEnv

        self.rootSecMethodIndex = None
        self.secMethodIndex = None
        self.secSystemIndex = None
        self.scfIterNr = 0

        self.equiv_atom_labels = []
        self.equiv_atom_pos = []


    def startedParsing(self, path, parser):
        """called when parsing starts"""
        self.parser = parser
        self.metaInfoEnv = self.parser.parserBuilder.metaInfoEnv

        # allows to reset values if the same superContext is used to parse different files
        self.initialize_values()

    def onClose_x_fleur_header(self, backend, gIndex, section):
        backend.addValue("program_version",
                         section["x_fleur_version"][0])


    def onOpen_section_system(self, backend, gIndex, section):
        self.equiv_atom_labels = []
        self.equiv_atom_pos = []

        mainFile = self.parser.fIn.fIn.name
        fName = mainFile[:-4] + ".inp"
        if os.path.exists(fName):
            structSuperContext = fleur_parser_inp.FleurInpContext()
            structParser = AncillaryParser(
                fileDescription = fleur_parser_inp.buildStructureMatchers(),
                parser = self.parser,
                cachingLevelForMetaName = fleur_parser_inp.get_cachingLevelForMetaName(self.metaInfoEnv, CachingLevel.PreOpenedIgnore),
                superContext = inpSuperContext)

            with open(fName) as fIn:
                inpParser.parseFile(fIn)



        mainFile = self.parser.fIn.fIn.name
        fName = mainFile[:-4] + ".xml"
        if os.path.exists(fName):
            xmlSuperContext = fleur_parser_xml.FleurXmlContext()
            xmlParser = AncillaryParser(
                fileDescription = fleur_parser_xml.buildXmlMatchers(),
                parser = self.parser,
                cachingLevelForMetaName = fleur_parser_xml.get_cachingLevelForMetaName(self.metaInfoEnv, CachingLevel.PreOpenedIgnore),
                superContext = xmlSuperContext)

            with open(fName) as fIn:
                xmlParser.parseFile(fIn)


        #if self.secSystemIndex is None:
        self.secSystemIndex = gIndex
        #        self.secSystemIndex["single_configuration_calculation_to_system_ref"] = gIndex


    def onClose_section_single_configuration_calculation(self, backend, gIndex, section):
        """Trigger called when section_single_configuration_calculation is opened.
        """
        # write number of SCF iterations
        backend.addValue('number_of_scf_iterations', self.scfIterNr)

        # write the references to section_method and section_system
        #        method_index = self.secMethodIndex("single_configuration_to_calculation_method_ref")
        #        if method_index is not None:

        backend.addValue('single_configuration_to_calculation_method_ref', self.secMethodIndex)
        #        system_index = self.secSystemIndex("single_configuration_calculation_to_system_ref")
        #        if system_index is not None:

        backend.addValue('single_configuration_calculation_to_system_ref', self.secSystemIndex)


    def onOpen_section_method(self, backend, gIndex, section):
        #if self.secMethodIndex is None:
        if self.rootSecMethodIndex is None:
            self.rootSecMethodIndex = gIndex
        self.secMethodIndex = gIndex
#        self.secMethodIndex["single_configuration_to_calculation_method_ref"] = gIndex

    def onClose_x_fleur_section_equiv_atoms(self, backend, gIndex, section):
        label = section["x_fleur_atom_name"][0]
        x = section["x_fleur_atom_pos_x"]
        y = section["x_fleur_atom_pos_y"]
        z = section["x_fleur_atom_pos_z"]
        if len(x) != len(y) or len(x) != len(z):
            raise Exception("incorrect parsing, different number of x,y,z components")
        groupPos = [[x[i], y[i], z[i]] for i in range(len(x))]
        nAt = len(groupPos)

        self.equiv_atom_labels += [label for i in range(nAt)]
        self.equiv_atom_pos += groupPos

    def onClose_section_system(self, backend, gIndex, section):

        #backend.addValue("smearing_kind", x_fleur_smearing_kind)
        smearing_kind = section['x_fleur_smearing_kind']
        if smearing_kind is not None:
        #    value = ''
            backend.addValue('x_fleur_smearing_kind', value)


        smearing_width = section['x_fleur_smearing_width']
        if smearing_width is not None:
        #    value = ''
            backend.addValue('x_fleur_smearing_width', value)



         #------1.atom_positions
        atom_pos = []
        for i in ['x', 'y', 'z']:
            api = section['x_fleur_atom_pos_' + i]
            if api is not None:
               atom_pos.append(api)
               logging.error("atom_pos: %s x %s y %s z %s",atom_pos, x, y, z)
        if atom_pos:
            # need to transpose array since its shape is [number_of_atoms,3] in the metadata
            backend.addArrayValues('atom_positions', np.transpose(np.asarray(atom_pos)), gIndex)
        elif len(self.equiv_atom_pos) > 0:
            backend.addArrayValues('atom_positions', np.asarray(self.equiv_atom_pos), gIndex)

         #------2.atom labels
        atom_labels = section['x_fleur_atom_name']
        if atom_labels is not None:
            backend.addArrayValues('atom_labels', np.asarray(atom_labels), gIndex)
        elif len(self.equiv_atom_labels) > 0:
            backend.addArrayValues('atom_labels', np.asarray(self.equiv_atom_labels), gIndex)

        #------3.atom force
        atom_force = []
        for i in ['x', 'y', 'z']:
            api = section['x_fleur_tot_for_' + i]
            if api is not None:
               atom_force.append(api)
        if atom_force:
            # need to transpose array since its shape is [number_of_atoms,3] in the metadata
           backend.addArrayValues('atom_forces', np.transpose(np.asarray(atom_force)))

        #----4. unit_cell
        unit_cell = []
        for i in ['x', 'y', 'z']:
            uci = section['x_fleur_lattice_vector_' + i]
            if uci is not None:
                unit_cell.append(uci)
        if unit_cell:
           backend.addArrayValues('simulation_cell', np.asarray(unit_cell))
           backend.addArrayValues("configuration_periodic_dimensions", np.ones(3, dtype=bool))


    def onClose_section_scf_iteration(self, backend, gIndex, section):
        #Trigger called when section_scf_iteration is closed.

        # count number of SCF iterations
        self.scfIterNr += 1

    def onClose_x_fleur_section_XC(self, backend, gIndex, section):
        xc_index = section["x_fleur_exch_pot"]
        logging.info("winsectxc: %s -> %s", section, xc_index)
        if not xc_index:
            xc_index = ["pbe"]
        xc_map_legend = {

            'pbe': ['GGA_X_PBE', 'GGA_C_PBE'],

            'rpbe': ['GGA_X_PBE', 'GGA_C_PBE'],

            'Rpbe': ['GGA_X_RPBE'],

            'pw91': ['GGA_X_PW91','GGA_C_PW91'],

            'l91': ['LDA_C_PW','LDA_C_PW_RPA','LDA_C_PW_MOD','LDA_C_OB_PW'],

            'vwn': ['LDA_C_VWN','LDA_C_VWN_1','LDA_C_VWN_2','LDA_C_VWN_3','LDA_C_VWN_4','LDA_C_VWN_RPA'],

            'bh': ['LDA_C_VBH'],

            'pz':['LDA_C_PZ'],

            'x-a': [],

            'mjw': [],

            'wign': [],

            'hl': [],

            'xal': [],

            'relativistic': ['---']
            #http://dx.doi.org/10.1088/0022-3719/12/15/007

        }

        # Push the functional string into the backend
        xc_map_legend = xc_map_legend.get(xc_index[0])
        if not xc_map_legend:
            raise Exception("Unhandled xc functional %s found" % xc_index)

        for xc_name in xc_map_legend:
            s = backend.openSection("section_XC_functionals")
            backend.addValue("XC_functional_name", xc_name)
            backend.closeSection("section_XC_functionals", s)



#*

#    def onClose_section_run(self, backend, gIndex, section):
#        """Trigger called when section_run is closed.
#        """
        # reset all variables
#        self.initialize_values()
#        backend.addValue("energy_total", energy_total)
#        # frame sequence
#        sampling_method = "geometry_optimization"

#        samplingGIndex = backend.openSection("section_sampling_method")
#        backend.addValue("sampling_method", sampling_method)
#        backend.closeSection("section_sampling_method", samplingGIndex)
#        frameSequenceGIndex = backend.openSection("section_frame_sequence")
#        backend.addValue("frame_sequence_to_sampling_ref", samplingGIndex)
#        backend.addArrayValues("frame_sequence_local_frames_ref", np.asarray(self.singleConfCalcs))
#        backend.closeSection("section_frame_sequence", frameSequenceGIndex)



"""    def onClose_x_fleur_section_xml_file(self, backend, gIndex, section):

        x_fleur_loading_xml_file_list = section['x_fleur_loading_xml_file']

        xml_file = x_fleur_loading_xml_file_list[-1]

        if xml_file is not None:
           logger.warning("This output showed this calculation need to load xml file, so we need this xml file ('%s') to read geometry information" % os.path.normpath(xml_file) )
           fName = os.path.normpath(xml_file)

           xmlSuperContext = FleurXMLParser.FleurXMLParserContext(False)
           xmlParser = AncillaryParser(
                fileDescription = FleurXMLParser.build_FleurXMLFileSimpleMatcher(),
                parser = self.parser,
                cachingLevelForMetaName = FleurXMLParser.get_cachingLevelForMetaName(self.metaInfoEnv, CachingLevel.Ignore),
                superContext = xmlSuperContext)

           try:
                with open(fName) as fxml:
                     xmlParser.parseFile(fxml)

           except IOError:
                logger.warning("Could not find xml file in directory '%s'. " % os.path.dirname(os.path.abspath(self.fName)))
"""

#####################################################################################################
#                                          description of the input                                 #
#####################################################################################################

mainFileDescription = SM(
              name = 'root',
              weak = True,
              startReStr = "",
              subMatchers = [
                SM(name = 'newRun',
                startReStr = r"\s* This output is generated by\s*[\w*.]+\s*\w*\*\s\*",
                repeats = True,
                required = True,
                forwardMatch = True,
                sections   = ['section_run','section_method', 'section_system', 'section_single_configuration_calculation'],
                subMatchers = [
### header ###
                    SM(name = 'header',
                    startReStr = r"\s* This output is generated by\s*(?P<x_fleur_version>[\w*.]+)\s*\w*\*\s\*",
                    sections=["x_fleur_header"],
                    fixedStartValues={'program_name': 'fleur', 'program_basis_set_type': '(L)APW+lo' }
                   ),

                    SM(name = 'systemName',
                    #startReStr = r"\s*-{11}fl7para file ends here-{11}\s*",#L112
                    startReStr = r"\s*[0-9]*\s*f l a p w  version\s[\w].*",
                    sections = ["section_system"],
                    subMatchers=[
                     #   SM(r"\s*[0-9]*\s*f l a p w  version\s[\w].*"),
                        SM(r"\s{4}(?P<x_fleur_system_name>[\w*\s*]+)\n"),#L117
                        SM(r"\s*name of space group=(?P<x_fleur_space_group>.*)"),#L121

                        SM(name = 'unit cell',
                        startReStr = r"\sbravais matrices of real and reciprocal lattices",
                        subMatchers=[
                            SM(r"\s*(?P<x_fleur_lattice_vector_x__bohr>[-+0-9.]+)\s+(?P<x_fleur_lattice_vector_y__bohr>[-+0-9.]+)\s+(?P<x_fleur_lattice_vector_z__bohr>[-+0-9.]+)\s*(?P<x_fleur_rec_lattice_vector_x__bohr>[-+0-9.]+)\s+(?P<x_fleur_rec_lattice_vector_y__bohr>[-+0-9.]+)\s+(?P<x_fleur_rec_lattice_vector_z__bohr>[-+0-9.]+)",#L131-5
                            repeats = True
                           )
                        ]),

                        SM(r"\s*the volume of the unit cell omega-tilda=\s*(?P<x_fleur_unit_cell_volume>[0-9.]+)"),#L137
                        SM(r"\s*the volume of the unit cell omega=\s*(?P<x_fleur_unit_cell_volume_omega>[0-9.]+)"), #L138

                        # SM(r"\s*exchange-correlation:\s*(?P<x_fleur_exch_pot>\w*\s*.*)",sections = ['x_fleur_section_XC']), #L140
                        SM(r"\s*exchange-correlation:\s*(?P<x_fleur_exch_pot>\w*)\s*(?P<x_fleur_xc_correction>\w*\s*.*)",sections = ['x_fleur_section_XC']), #L140

                        SM(name = 'atomPositions',
                        startReStr = r"\s*(?P<x_fleur_atom_name>\w*)\s+(?P<x_fleur_nuclear_number>[0-9]+)\s+(?P<x_fleur_number_of_core_levels>[0-9]+)\s+(?P<x_fleur_lexpansion_cutoff>[0-9.]+)\s+(?P<x_fleur_mt_gridpoints>[0-9.]+)\s+(?P<x_fleur_mt_radius>[0-9.]+)\s+(?P<x_fleur_logarythmic_increment>[0-9.]+)",
                           sections=["x_fleur_section_equiv_atoms"],
                        repeats = True,
                           subMatchers=[
                               SM(name = 'equiv atoms',
                               startReStr = r"\s*(?P<x_fleur_atom_pos_x__bohr>[-+0-9.]+)\s+(?P<x_fleur_atom_pos_y__bohr>[-+0-9.]+)\s+(?P<x_fleur_atom_pos_z__bohr>[-+0-9.]+)\s+(?P<x_fleur_atom_coord_scale>[-+0-9.]*)",
                        #       sections=["x_fleur_section_equiv_atoms"],
                                  subMatchers = [
                                      SM(r"\s*(?P<x_fleur_atom_pos_x__bohr>[-+0-9.]+)\s+(?P<x_fleur_atom_pos_y__bohr>[-+0-9.]+)\s+(?P<x_fleur_atom_pos_z__bohr>[-+0-9.]+)\s+(?P<x_fleur_atom_coord_scale>[-+0-9.]*)",#L145
                                      repeats = True
                                     )
                                  ]
                              )
                           ]),


                        SM(r"\s* Suggested values for input:"),
                        SM(r"\s*k_max\s=\s*(?P<x_fleur_k_max>.*)"),#L154
                        SM(r"\s*G_max\s=\s*(?P<x_fleur_G_max>.*)"),#L155
                        SM(r"\s*volume of interstitial region=\s*(?P<x_fleur_vol_interstitial>[0-9.]+)"),#L157
                        SM(r"\s*number of atom types=\s*(?P<x_fleur_nr_of_atom_types>[0-9]+)"),#L160
                        SM(r"\s*total number of atoms=\s*(?P<x_fleur_total_atoms>[0-9]+)"),

                        SM(r"\s*(?P<x_fleur_smearing_kind>\w*)-integration is used\s*.*"),#L187
                        SM(r"\s*gaussian half width\s*=\s*(?P<x_fleur_smearing_width__hartree>[0-9.]+)"),#188
                        SM(r"\s*number of valence electrons=\s*(?P<x_fleur_nr_of_valence_electrons>[0-9.]+)"),#190
                        SM(r"\s*temperature broadening\s*=\s*(?P<x_fleur_smearing_temperature>[0-9.]+)"),#191

                        SM(r"\s*number of k-points for this window =\s*(?P<x_fleur_nkptd>[0-9]*)"), #723
                        SM(name = 'k-point',
                           startReStr = r"\s{11}coordinates\s{11}weights",
                           subMatchers = [
                               SM(r"\s+(?P<x_fleur_k_point_x>[-+0-9.]+)\s+(?P<x_fleur_k_point_y>[-+0-9.]+)\s+(?P<x_fleur_k_point_z>[-+0-9.]+)\s+(?P<x_fleur_k_point_weight>[-+0-9.]+)", repeats = True   #L725
                              )
                           ]),
                        SM(r"\s*total electronic charge   =\s*(?P<x_fleur_tot_elec_charge>.*)"),#L1107
                        SM(r"\s*total nuclear charge      =\s*(?P<x_fleur_tot_nucl_charge>.*)") #L1108

                    ]),
                    #Check problematic SM
                    SM(name = "scf iteration",
                       startReStr = r"\s*it=\s*(?P<x_fleur_iteration_number>[0-9]+)",
                       sections=["section_scf_iteration"],
                       repeats = True,
                       subMatchers=[
                           SM(r"\s*---->\s*total energy=\s*(?P<x_fleur_energy_total__hartree>[-+0-9.]+)\s*htr"),


                           SM(startReStr = r"\s*\W{5}\s+TOTAL FORCES ON ATOMS\s+\W{5}",
                              subMatchers = [
                                  SM(r"\sTOTAL FORCE FOR ATOM TYPE=\s*[0-9]\s+X=\s+(?P<x_fleur_tot_for_x>[0-9.]+)\s+Y=\s+(?P<x_fleur_tot_for_y>[0-9.]+)\s+Z=\s+(?P<x_fleur_tot_for_z>[0-9.]+)", repeats = True),#L3825 first
                                  SM(r"\s*FX_TOT=(?P<x_fleur_tot_for_fx>[-+0-9.]+)\sFY_TOT=(?P<x_fleur_tot_for_fy>[-+0-9.]+)\sFZ_TOT=(?P<x_fleur_tot_for_fz>[-+0-9.]+)", repeats = True)
                              ], repeats = True),


                           SM(r"\s*---->\s*.*tkb\*entropy.*=\s*(?P<x_fleur_entropy__hartree>[-+0-9.]+)\shtr"),
                           SM(r"\s*---->\s*free energy=\s*(?P<x_fleur_free_energy__hartree>[-+0-9.]+)\shtr")
                       ]
                   )
                ])
              ])

# which values to cache or forward (mapping meta name -> CachingLevel)

cachingLevelForMetaName = {

    "XC_functional_name": CachingLevel.ForwardAndCache,
#    "energy_total": CachingLevel.ForwardAndCache,


 }
# loading metadata from nomad-meta-info/meta_info/nomad_meta_info/fleur.nomadmetainfo.json

parserInfo = {
  "name": "Fleur_parser",
  "version": "1.0"
}

class FleurParser():
    """ A proper class envolop for running this parser from within python. """
    def __init__(self, backend, **kwargs):
        self.backend_factory = backend

    def parse(self, mainfile):
        from unittest.mock import patch
        logging.debug('fleur parser started')
        logging.getLogger('nomadcore').setLevel(logging.WARNING)
        backend = self.backend_factory("fleur.nomadmetainfo.json")
        with patch.object(sys, 'argv', ['<exe>', '--uri', 'nmd://uri', mainfile]):
            mainFunction(
                mainFileDescription=mainFileDescription,
                metaInfoEnv=None,
                parserInfo=parserInfo,
                superContext=FleurContext(),
                superBackend=backend)

        return backend



# pylint: disable=relative-beyond-top-level,import-outside-toplevel
import unittest
from typing import Union, List, Set

from optimade.server.config import CONFIG
from optimade.server import mappers
from optimade.server.entry_collections import CI_FORCE_MONGO

from .utils import SetClient

MONGOMOCK_OLD = False
MONGOMOCK_MSG = ""
if not CI_FORCE_MONGO and not CONFIG.use_real_mongo:
    import mongomock

    MONGOMOCK_OLD = tuple(
        int(val) for val in mongomock.__version__.split(".")[0:3]
    ) <= (3, 19, 0)
    MONGOMOCK_MSG = f"mongomock version {mongomock.__version__}<=3.19.0 is too old for this test, skipping..."


class IncludeTests(SetClient, unittest.TestCase):
    """Make sure `include` is handled correctly

    NOTE: Currently _only_ structures have relationships (references).
    """

    server = "regular"

    def _check_response(
        self,
        request: str,
        expected_included_types: Union[List, Set],
        expected_included_resources: Union[List, Set],
        expected_relationship_types: Union[List, Set] = None,
    ):
        try:
            response = self.client.get(request)
            self.assertEqual(
                response.status_code, 200, msg=f"Request failed: {response.json()}"
            )

            response = response.json()
            response_data = (
                response["data"]
                if isinstance(response["data"], list)
                else [response["data"]]
            )

            included_resource_types = list({_["type"] for _ in response["included"]})
            self.assertEqual(
                sorted(expected_included_types),
                sorted(included_resource_types),
                msg=f"Expected relationship types: {expected_included_types}. "
                f"Does not match relationship types in response's included field: {included_resource_types}",
            )

            if expected_relationship_types is None:
                expected_relationship_types = expected_included_types
            relationship_types = set()
            for entry in response_data:
                relationship_types.update(set(entry.get("relationships", {}).keys()))
            self.assertEqual(
                sorted(expected_relationship_types),
                sorted(relationship_types),
                msg=f"Expected relationship types: {expected_relationship_types}. "
                f"Does not match relationship types found in response data: {relationship_types}",
            )

            included_resources = [_["id"] for _ in response["included"]]
            self.assertEqual(
                len(included_resources),
                len(expected_included_resources),
                msg=response["included"],
            )
            self.assertEqual(
                sorted(set(included_resources)), sorted(expected_included_resources)
            )

        except Exception as exc:
            print("Request attempted:")
            print(f"{self.client.base_url}{request}")
            raise exc

    def _check_error_response(
        self,
        request: str,
        expected_status: int = None,
        expected_title: str = None,
        expected_detail: str = None,
    ):
        expected_status = 400 if expected_status is None else expected_status
        expected_title = "Bad Request" if expected_title is None else expected_title
        super()._check_error_response(
            request, expected_status, expected_title, expected_detail
        )

    def test_default_value(self):
        """Default value for `include` is 'references'

        Test also that passing `include=` equals passing the default value
        """
        request = "/structures"
        expected_types = ["references"]
        expected_reference_ids = ["dijkstra1968", "maddox1988", "dummy/2019"]
        self._check_response(request, expected_types, expected_reference_ids)

    def test_empty_value(self):
        """An empty value should resolve in no relationships being returned under `included`"""
        request = "/structures?include="
        expected_types = []
        expected_reference_ids = []
        expected_data_relationship_types = ["references"]
        self._check_response(
            request,
            expected_types,
            expected_reference_ids,
            expected_data_relationship_types,
        )

    def test_default_value_single_entry(self):
        """For single entry. Default value for `include` is 'references'"""
        request = "/structures/mpf_1"
        expected_types = ["references"]
        expected_reference_ids = ["dijkstra1968"]
        self._check_response(request, expected_types, expected_reference_ids)

    def test_empty_value_single_entry(self):
        """For single entry. An empty value should resolve in no relationships being returned under `included`"""
        request = "/structures/mpf_1?include="
        expected_types = []
        expected_reference_ids = []
        expected_data_relationship_types = ["references"]
        self._check_response(
            request,
            expected_types,
            expected_reference_ids,
            expected_data_relationship_types,
        )

    def test_wrong_relationship_type(self):
        """A wrong type should result in a `400 Bad Request` response"""
        from optimade.server.routers import ENTRY_COLLECTIONS

        for wrong_type in ("test", '""', "''"):
            request = f"/structures?include={wrong_type}"
            error_detail = (
                f"'{wrong_type}' cannot be identified as a valid relationship type. "
                f"Known relationship types: {sorted(ENTRY_COLLECTIONS.keys())}"
            )
            self._check_error_response(request, expected_detail=error_detail)


class ResponseFieldTests(SetClient, unittest.TestCase):
    """Make sure response_fields is handled correctly"""

    server = "regular"

    get_mapper = {
        "links": mappers.LinksMapper,
        "references": mappers.ReferenceMapper,
        "structures": mappers.StructureMapper,
    }

    def required_fields_test_helper(
        self, endpoint: str, known_unused_fields: set, expected_fields: set
    ):
        """Utility function for creating required fields tests"""
        expected_fields |= (
            self.get_mapper[endpoint].get_required_fields() - known_unused_fields
        )
        expected_fields.add("attributes")
        request = f"/{endpoint}?response_fields={','.join(expected_fields)}"

        # Check response
        try:
            response = self.client.get(request)
            self.assertEqual(
                response.status_code, 200, msg=f"Request failed: {response.json()}"
            )

            response = response.json()
            response_fields = set()
            for entry in response["data"]:
                response_fields.update(set(entry.keys()))
                response_fields.update(set(entry["attributes"].keys()))
            self.assertEqual(sorted(expected_fields), sorted(response_fields))
        except Exception as exc:
            print("Request attempted:")
            print(f"{self.client.base_url}{request}")
            raise exc

    def test_required_fields_links(self):
        """Certain fields are REQUIRED, no matter the value of `response_fields`"""
        endpoint = "links"
        illegal_top_level_field = "relationships"
        non_used_top_level_fields = {"links"}
        non_used_top_level_fields.add(illegal_top_level_field)
        expected_fields = {"homepage", "base_url"}
        self.required_fields_test_helper(
            endpoint, non_used_top_level_fields, expected_fields
        )

    def test_required_fields_references(self):
        """Certain fields are REQUIRED, no matter the value of `response_fields`"""
        endpoint = "references"
        non_used_top_level_fields = {"links", "relationships"}
        expected_fields = {"year", "journal"}
        self.required_fields_test_helper(
            endpoint, non_used_top_level_fields, expected_fields
        )

    def test_required_fields_structures(self):
        """Certain fields are REQUIRED, no matter the value of `response_fields`"""
        endpoint = "structures"
        non_used_top_level_fields = {"links"}
        expected_fields = {"elements", "nelements"}
        self.required_fields_test_helper(
            endpoint, non_used_top_level_fields, expected_fields
        )


class FilterTests(SetClient, unittest.TestCase):

    server = "regular"

    def test_custom_field(self):
        request = '/structures?filter=_exmpl_chemsys="Ac"'
        expected_ids = ["mpf_1"]
        self._check_response(request, expected_ids, len(expected_ids))

    def test_id(self):
        request = "/structures?filter=id=mpf_2"
        expected_ids = ["mpf_2"]
        self._check_response(request, expected_ids, len(expected_ids))

    def test_geq(self):
        request = "/structures?filter=nelements>=9"
        expected_ids = ["mpf_3819"]
        self._check_response(request, expected_ids, len(expected_ids))

    def test_gt(self):
        request = "/structures?filter=nelements>8"
        expected_ids = ["mpf_3819"]
        self._check_response(request, expected_ids, len(expected_ids))

    def test_rhs_comparison(self):
        request = "/structures?filter=8<nelements"
        expected_ids = ["mpf_3819"]
        self._check_response(request, expected_ids, len(expected_ids))

    def test_gt_none(self):
        request = "/structures?filter=nelements>9"
        expected_ids = []
        self._check_response(request, expected_ids, len(expected_ids))

    def test_list_has(self):
        request = '/structures?filter=elements HAS "Ti"'
        expected_ids = ["mpf_3803", "mpf_3819"]
        self._check_response(request, expected_ids, len(expected_ids))

    def test_page_limit(self):
        request = '/structures?filter=elements HAS "Ac"&page_limit=2'
        expected_ids = ["mpf_1", "mpf_2"]
        expected_return = 6
        self._check_response(request, expected_ids, expected_return)

        request = '/structures?page_limit=2&filter=elements HAS "Ac"'
        expected_ids = ["mpf_1", "mpf_2"]
        expected_return = 6
        self._check_response(request, expected_ids, expected_return)

    def test_page_limit_max(self):
        request = f"/structures?page_limit={CONFIG.page_limit_max + 1}"
        self._check_error_response(
            request,
            expected_status=403,
            expected_title="HTTPException",
            expected_detail=f"Max allowed page_limit is {CONFIG.page_limit_max}, you requested {CONFIG.page_limit_max + 1}",
        )

    def test_value_list_operator(self):
        request = "/structures?filter=dimension_types HAS < 1"
        self._check_error_response(
            request,
            expected_status=501,
            expected_title="NotImplementedError",
            expected_detail="set_op_rhs not implemented for use with OPERATOR. Given: [Token(HAS, 'HAS'), Token(OPERATOR, '<'), 1]",
        )

    def test_has_any_operator(self):
        request = "/structures?filter=dimension_types HAS ANY > 1"
        self._check_error_response(
            request,
            expected_status=501,
            expected_title="NotImplementedError",
            expected_detail="OPERATOR > inside value_list [Token(OPERATOR, '>'), 1] not implemented.",
        )

    def test_list_has_all(self):
        request = '/structures?filter=elements HAS ALL "Ba","F","H","Mn","O","Re","Si"'
        expected_ids = ["mpf_3819"]
        self._check_response(request, expected_ids, len(expected_ids))

        request = '/structures?filter=elements HAS ALL "Re","Ti"'
        expected_ids = ["mpf_3819"]
        self._check_response(request, expected_ids, len(expected_ids))

    def test_list_has_any(self):
        request = '/structures?filter=elements HAS ANY "Re","Ti"'
        expected_ids = ["mpf_3819", "mpf_3803"]
        self._check_response(request, expected_ids, len(expected_ids))

    def test_list_length_basic(self):
        request = "/structures?filter=elements LENGTH = 9"
        expected_ids = ["mpf_3819"]
        self._check_response(request, expected_ids, len(expected_ids))

        request = "/structures?filter=elements LENGTH 9"
        self._check_response(request, expected_ids, len(expected_ids))

    def test_list_length(self):
        request = "/structures?filter=elements LENGTH >= 9"
        expected_ids = ["mpf_3819"]
        self._check_response(request, expected_ids, len(expected_ids))

        request = "/structures?filter=structure_features LENGTH > 0"
        expected_ids = []
        self._check_response(request, expected_ids, len(expected_ids))

        request = "/structures?filter=structure_features LENGTH > 0"
        expected_ids = []
        self._check_response(request, expected_ids, len(expected_ids))

        request = "/structures?filter=cartesian_site_positions LENGTH > 43"
        expected_ids = ["mpf_551", "mpf_3803", "mpf_3819"]
        self._check_response(request, expected_ids, len(expected_ids))

        request = "/structures?filter=species_at_sites LENGTH > 43"
        expected_ids = ["mpf_551", "mpf_3803", "mpf_3819"]
        self._check_response(request, expected_ids, len(expected_ids))

        request = "/structures?filter=nsites LENGTH > 43"
        expected_ids = []
        self._check_response(request, expected_ids, len(expected_ids))

        request = "/structures?filter=structure_features LENGTH != 0"
        error_detail = "Operator != not implemented for LENGTH filter."
        self._check_error_response(
            request,
            expected_status=501,
            expected_title="NotImplementedError",
            expected_detail=error_detail,
        )

    @unittest.skipIf(MONGOMOCK_OLD, MONGOMOCK_MSG)
    def test_list_has_only(self):
        """ Test HAS ONLY query on elements.

        This test fails with mongomock<=3.19.0 when $size is 1, but works with a real mongo.

        TODO: this text and skip condition should be removed once mongomock>3.19.0 has been released, which should
        contain the bugfix for this: https://github.com/mongomock/mongomock/pull/597.

        """

        request = '/structures?filter=elements HAS ONLY "Ac", "Mg"'
        expected_ids = ["mpf_23"]
        self._check_response(request, expected_ids, len(expected_ids))

        request = '/structures?filter=elements HAS ONLY "Ac"'
        expected_ids = ["mpf_1"]
        self._check_response(request, expected_ids, len(expected_ids))

    def test_list_correlated(self):
        request = '/structures?filter=elements:elements_ratios HAS "Ag":"0.2"'
        self._check_error_response(
            request, expected_status=501, expected_title="NotImplementedError"
        )
        # expected_ids = ["mpf_259"]
        # self._check_response(request, expected_ids, len(expected_ids))

    def test_is_known(self):
        request = "/structures?filter=nsites IS KNOWN AND nsites>=44"
        expected_ids = ["mpf_551", "mpf_3803", "mpf_3819"]
        self._check_response(request, expected_ids, len(expected_ids))

        request = "/structures?filter=lattice_vectors IS KNOWN AND nsites>=44"
        expected_ids = ["mpf_551", "mpf_3803", "mpf_3819"]
        self._check_response(request, expected_ids, len(expected_ids))

    def test_aliased_is_known(self):
        request = "/structures?filter=id IS KNOWN AND nsites>=44"
        expected_ids = ["mpf_551", "mpf_3803", "mpf_3819"]
        self._check_response(request, expected_ids, len(expected_ids))

        request = "/structures?filter=chemical_formula_reduced IS KNOWN AND nsites>=44"
        expected_ids = ["mpf_551", "mpf_3803", "mpf_3819"]
        self._check_response(request, expected_ids, len(expected_ids))

        request = (
            "/structures?filter=chemical_formula_descriptive IS KNOWN AND nsites>=44"
        )
        expected_ids = ["mpf_551", "mpf_3803", "mpf_3819"]
        self._check_response(request, expected_ids, len(expected_ids))

    def test_aliased_fields(self):
        request = '/structures?filter=chemical_formula_anonymous="A"'
        expected_ids = ["mpf_1", "mpf_200"]
        self._check_response(request, expected_ids, len(expected_ids))

        request = '/structures?filter=chemical_formula_anonymous CONTAINS "A2BC"'
        expected_ids = ["mpf_2", "mpf_3", "mpf_110"]
        self._check_response(request, expected_ids, len(expected_ids))

    def test_string_contains(self):
        request = '/structures?filter=chemical_formula_descriptive CONTAINS "c2Ag"'
        expected_ids = ["mpf_3", "mpf_2"]
        self._check_response(request, expected_ids, len(expected_ids))

    def test_string_start(self):
        request = (
            '/structures?filter=chemical_formula_descriptive STARTS WITH "Ag2CSNCl"'
        )
        expected_ids = ["mpf_259"]
        self._check_response(request, expected_ids, len(expected_ids))

    def test_string_end(self):
        request = '/structures?filter=chemical_formula_descriptive ENDS WITH "NClO4"'
        expected_ids = ["mpf_259"]
        self._check_response(request, expected_ids, len(expected_ids))

    def test_list_has_and(self):
        request = '/structures?filter=elements HAS "Ac" AND nelements=1'
        expected_ids = ["mpf_1"]
        self._check_response(request, expected_ids, len(expected_ids))

    def test_awkward_not_queries(self):
        """ Test an awkward query from the spec examples. It should return all but 2 structures
        in the test data. The test is done in three parts:

            - first query the individual expressions that make up the OR,
            - then do an empty query to get all IDs
            - then negate the expressions and ensure that all IDs are returned except
              those from the first queries.

        """
        expected_ids = ["mpf_3819"]
        request = (
            '/structures?filter=chemical_formula_descriptive="Ba2NaTi2MnRe2Si8HO26F" AND '
            'chemical_formula_anonymous = "A26B8C2D2E2FGHI" '
        )
        self._check_response(request, expected_ids, len(expected_ids))

        expected_ids = ["mpf_2"]
        request = (
            '/structures?filter=chemical_formula_anonymous = "A2BC" AND '
            'NOT chemical_formula_descriptive = "Ac2AgPb" '
        )
        self._check_response(request, expected_ids, len(expected_ids))

        request = "/structures"
        unexpected_ids = ["mpf_3819", "mpf_2"]
        expected_ids = [
            structure["id"]
            for structure in self.client.get(request).json()["data"]
            if structure["id"] not in unexpected_ids
        ]

        request = (
            "/structures?filter="
            "NOT ( "
            'chemical_formula_descriptive = "Ba2NaTi2MnRe2Si8HO26F" AND '
            'chemical_formula_anonymous = "A26B8C2D2E2FGHI" OR '
            'chemical_formula_anonymous = "A2BC" AND '
            'NOT chemical_formula_descriptive = "Ac2AgPb" '
            ")"
        )
        self._check_response(request, expected_ids, len(expected_ids))

    def test_not_or_and_precedence(self):
        request = '/structures?filter=NOT elements HAS "Ac" AND nelements=1'
        expected_ids = ["mpf_200"]
        self._check_response(request, expected_ids, len(expected_ids))

        request = '/structures?filter=nelements=1 AND NOT elements HAS "Ac"'
        expected_ids = ["mpf_200"]
        self._check_response(request, expected_ids, len(expected_ids))

        request = '/structures?filter=NOT elements HAS "Ac" AND nelements=1 OR nsites=1'
        expected_ids = ["mpf_1", "mpf_200"]
        self._check_response(request, expected_ids, len(expected_ids))

        request = '/structures?filter=elements HAS "Ac" AND nelements>1 AND nsites=1'
        expected_ids = []
        self._check_response(request, expected_ids, len(expected_ids))

    def test_brackets(self):
        request = '/structures?filter=elements HAS "Ac" AND nelements=1 OR nsites=1'
        expected_ids = ["mpf_200", "mpf_1"]
        self._check_response(request, expected_ids, len(expected_ids))

        request = '/structures?filter=(elements HAS "Ac" AND nelements=1) OR (elements HAS "Ac" AND nsites=1)'
        expected_ids = ["mpf_1"]
        self._check_response(request, expected_ids, len(expected_ids))

    def test_filter_on_relationships(self):
        request = '/structures?filter=references.id HAS "dummy/2019"'
        expected_ids = ["mpf_3819"]
        self._check_response(request, expected_ids, len(expected_ids))

        request = (
            '/structures?filter=references.id HAS ANY "dummy/2019", "dijkstra1968"'
        )
        expected_ids = ["mpf_1", "mpf_2", "mpf_3819"]
        self._check_response(request, expected_ids, len(expected_ids))

        request = '/structures?filter=references.id HAS ONLY "dijkstra1968"'
        expected_ids = ["mpf_1", "mpf_2"]
        self._check_response(request, expected_ids, len(expected_ids))

        request = '/structures?filter=references.doi HAS ONLY "10/123"'
        error_detail = (
            'Cannot filter relationships by field "doi", only "id" is supported.'
        )
        self._check_error_response(
            request,
            expected_status=501,
            expected_title="NotImplementedError",
            expected_detail=error_detail,
        )

    def _check_response(
        self, request: str, expected_ids: Union[List, Set], expected_return: int
    ):
        try:
            response = self.client.get(request)
            self.assertEqual(
                response.status_code, 200, msg=f"Request failed: {response.json()}"
            )
            response = response.json()
            response_ids = [struct["id"] for struct in response["data"]]
            self.assertEqual(sorted(expected_ids), sorted(response_ids))
            self.assertEqual(response["meta"]["data_returned"], expected_return)
        except Exception as exc:
            print("Request attempted:")
            print(f"{self.client.base_url}{request}")
            raise exc

    def _check_error_response(
        self,
        request: str,
        expected_status: int = None,
        expected_title: str = None,
        expected_detail: str = None,
    ):
        expected_status = 500 if expected_status is None else expected_status
        super()._check_error_response(
            request, expected_status, expected_title, expected_detail
        )



# Copyright 2018 Markus Scheidgen
#
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict
import numpy as np
import re
import ase.data
from string import ascii_uppercase
import pint.quantity

from nomad.normalizing.normalizer import SystemBasedNormalizer
from nomad.units import ureg
from nomad.datamodel import OptimadeEntry, Species, DFTMetadata, EntryMetadata
from nomad.datamodel.metainfo.public import section_system

species_re = re.compile(r'^([A-Z][a-z]?)(\d*)$')


class OptimadeNormalizer(SystemBasedNormalizer):

    '''
    This normalizer performs all produces a section all data necessary for the Optimade API.
    It assumes that the :class:`SystemNormalizer` was run before.
    '''
    def __init__(self, archive):
        super().__init__(archive, only_representatives=True)

    def add_optimade_data(self, index) -> OptimadeEntry:
        '''
        The 'main' method of this :class:`SystemBasedNormalizer`.
        Normalizes the section with the given `index`.
        Normalizes geometry, classifies, system_type, and runs symmetry analysis.
        '''

        if self.entry_archive.section_metadata is None:
            self.entry_archive.m_create(EntryMetadata)
        if self.entry_archive.section_metadata.dft is None:
            self.entry_archive.section_metadata.m_create(DFTMetadata)
        optimade = self.entry_archive.section_metadata.dft.m_create(OptimadeEntry)

        def get_value(quantity_def, default: Any = None, numpy: bool = False, unit=None) -> Any:
            try:
                value = self.section_run.section_system[-1].m_get(quantity_def)
                if type(value) == np.ndarray and not numpy:
                    return value.tolist()
                if isinstance(value, list) and numpy:
                    return np.array(value)

                if numpy and unit is not None:
                    if isinstance(value, pint.quantity._Quantity):
                        value = value.to(unit)
                    else:
                        value = value * unit

                return value
            except KeyError:
                return default

        from nomad.normalizing.system import normalized_atom_labels

        nomad_species = get_value(section_system.atom_labels)

        # elements
        atoms = normalized_atom_labels(nomad_species)
        atom_count = len(atoms)
        atom_counts: Dict[str, int] = {}
        for atom in atoms:
            current = atom_counts.setdefault(atom, 0)
            current += 1
            atom_counts[atom] = current

        optimade.elements = list(set(atoms))
        optimade.elements.sort()
        optimade.nelements = len(optimade.elements)
        optimade.elements_ratios = [
            atom_counts[element] / atom_count
            for element in optimade.elements]

        # formulas
        optimade.chemical_formula_reduced = get_value(section_system.chemical_composition_reduced)
        optimade.chemical_formula_hill = get_value(section_system.chemical_composition_bulk_reduced)
        optimade.chemical_formula_descriptive = optimade.chemical_formula_hill
        optimade.chemical_formula_anonymous = ''
        for i in range(len(optimade.elements)):
            part = '%s' % ascii_uppercase[i % len(ascii_uppercase)]
            if atom_counts[optimade.elements[i]] > 1:
                part += str(atom_counts[optimade.elements[i]])
            optimade.chemical_formula_anonymous += part

        # sites
        optimade.nsites = len(nomad_species)
        optimade.species_at_sites = nomad_species
        optimade.lattice_vectors = get_value(section_system.lattice_vectors, numpy=True, unit=ureg.m)
        optimade.cartesian_site_positions = get_value(section_system.atom_positions, numpy=True, unit=ureg.m)
        optimade.dimension_types = [
            1 if value else 0
            for value in get_value(section_system.configuration_periodic_dimensions)]

        # species
        for species_label in set(nomad_species):
            match = re.match(species_re, species_label)

            element_label = match.group(1)

            species = optimade.m_create(Species)
            species.name = species_label
            if element_label in ase.data.chemical_symbols:
                chemical_symbol = element_label
            else:
                chemical_symbol = 'x'
            species.chemical_symbols = [chemical_symbol]
            species.concentration = [1.0]

        optimade.structure_features = []

        return optimade

    def normalize_system(self, system, is_representative):
        if not is_representative:
            return False

        try:
            self.add_optimade_data(system.m_parent_index)
            return True

        except Exception as e:
            self.logger.warn('could not acquire optimade data', exc_info=e)



import numpy as np            # pylint: disable=unused-import
import typing                 # pylint: disable=unused-import
from nomad.metainfo import (  # pylint: disable=unused-import
    MSection, MCategory, Category, Package, Quantity, Section, SubSection, SectionProxy,
    Reference
)
from nomad.metainfo.legacy import LegacyDefinition

from nomad.datamodel.metainfo import public

m_package = Package(
    name='phonopy_nomadmetainfo_json',
    description='None',
    a_legacy=LegacyDefinition(name='phonopy.nomadmetainfo.json'))


class x_phonopy_input(MCategory):
    '''
    Information about properties that concern phonopy calculations.
    '''

    m_def = Category(
        a_legacy=LegacyDefinition(name='x_phonopy_input'))


class section_method(public.section_method):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_method'))

    x_phonopy_displacement = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='meter',
        description='''
        Amplitude of the atom diplacement for the phonopy supercell
        ''',
        categories=[x_phonopy_input],
        a_legacy=LegacyDefinition(name='x_phonopy_displacement'))

    x_phonopy_symprec = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='meter',
        description='''
        Symmetry threshold for the space group identification of the crystal for which the
        vibrational properties are to be calculated
        ''',
        categories=[x_phonopy_input],
        a_legacy=LegacyDefinition(name='x_phonopy_symprec'))


class section_system(public.section_system):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_system'))

    x_phonopy_original_system_ref = Quantity(
        type=public.section_system,
        shape=[],
        description='''
        Original cell from which the supercell for the DFT calculations was constructed
        ''',
        a_legacy=LegacyDefinition(name='x_phonopy_original_system_ref'))


m_package.__init_metainfo__()
